#!/usr/bin/python
# -*- coding: utf-8 -*-

from updatergui import UpdaterGUI
import Tkinter as tk
import tkMessageBox
from i18n import _
import hid
from hidapiwrapper import HidApiWrapper
from updatereader import UpdateReader
import deviceprogrammer
import time
from mikroeuhb.device import Device
from mikroeuhb.device import ExpectException
import os
import glob
import sys
import traceback
import slideshow
import winregistry

def windows_version():
	if sys.platform != 'win32':
		return (0, 0)
	winversion = os.sys.getwindowsversion()
	return (winversion[0], winversion[1])

def is_xp():
	major, minor = windows_version()
	return major == 5 and minor > 0

def is_7_or_vista():
	major, minor = windows_version()
	return major == 6 and minor < 2

def system_needs_firsttime():
	return is_xp() or is_7_or_vista()

class UpdateState:
	INITIAL = 1
	LOADED = 2
	WAITING_FOR_DEVICE = 3
	UPDATING = 4

class AodyoUpdate:
	OLD_VENDOR_ID = 0x1234
	OLD_PRODUCT_ID = 0x0001
	VENDOR_ID = 0x16c0
	PRODUCT_ID = 0x05df
	UPDATE_TEXT = _(fr=u'Mettre à jour', en=u'Update')
	CANCEL_TEXT = _(fr=u'Annuler', en=u'Cancel')
	STRFTIME_PATTERN = _(fr="%d/%m/%Y", en="%Y-%m-%d")
	TITLE = _(fr='Mise à jour Aodyo', en='Aodyo Update')
	BEFOREUPDATE_INFO = _(fr=u"Ne branchez pas tout de suite votre %s.\nCliquez sur le bouton \"%s\" pour commencer la mise à jour.", en=u"Do not plug your %s in right now.\nClick on \"%s\" to start the update process.")

	def __init__(self, filename = None):
		self.first_time_done = True
		self.needs_firsttime = system_needs_firsttime()
		if self.needs_firsttime:
			self.first_time_done = winregistry.already_plugged_in([(AodyoUpdate.OLD_VENDOR_ID, AodyoUpdate.OLD_PRODUCT_ID), (AodyoUpdate.VENDOR_ID, AodyoUpdate.PRODUCT_ID)])
		self.failure_reason = None
		self.update_is_bundled = False
		self.root = tk.Tk()
		self.root.title(AodyoUpdate.TITLE)
		self.gui = UpdaterGUI(self.root, self.onupdateclick, self.onload)
		self.gui.set_update_btn_enabled(False)
		self.gui.set_load_btn_enabled(True)
		self.gui.set_product_text(AodyoUpdate.TITLE)
		self.gui.set_info_text(_(fr=u"Veuillez ouvrir un fichier de mise à jour avant de continuer.", en=u"Please load an update file before continuing."))
		if sys.platform == 'darwin':
			self.root.createcommand("::tk::mac::OpenDocument", self.respond_to_OpenDocument)
		if filename:
			self.onload(filename, True)
		else:
			self.look_for_bundled_update()

	def mainloop(self):
		if sys.platform == 'darwin' and hasattr(sys, '_MEIPASS'):
			# os.system('''/usr/bin/osascript -e 'tell application "System Events" to keystroke "	" using {command down, shift down}' ''')
			self.root.lift()
		self.root.mainloop()

	def respond_to_OpenDocument(self, *args):
		if len(args) > 0:
			self.onload(args[0], False)

	def look_for_recent_update(self):
		wholeglob = '*.aodyoupdate'

		# If we're in a pyinstaller-generated executable, check its folder
		if getattr(sys, 'frozen', False):
			execdir = os.path.dirname(sys.executable)
			if sys.platform == 'darwin':
				execdir = os.path.dirname(os.path.dirname(os.path.dirname(execdir)))
			wholeglob = os.path.join(execdir, wholeglob)

		recentfiles = glob.iglob(wholeglob)
		try:
			recentfile = max(recentfiles, key=os.path.getctime)
			if recentfile:
				self.onload(recentfile, True)
		except:
			pass

	def look_for_bundled_update(self):
		wholeglob = '*.aodyoupdate'

		# If we're in a pyinstaller-generated executable, check its folder
		if getattr(sys, 'frozen', False):
			execdir = os.path.dirname(sys.executable)
			if sys.platform == 'darwin':
				execdir = os.path.join(os.path.dirname(execdir), 'Resources')
			elif sys.platform == 'win32' and hasattr(sys, '_MEIPASS'):
				execdir = sys._MEIPASS
			wholeglob = os.path.join(execdir, wholeglob)

		recentfiles = glob.iglob(wholeglob)
		try:
			recentfile = max(recentfiles, key=os.path.getctime)
			if recentfile:
				self.update_is_bundled = True
				self.onload(recentfile, True)
			else:
				self.update_is_bundled = False
				self.look_for_recent_update()
		except:
			pass

	def load_update_info(self, update):
		self.update = update
		datestr = time.strftime(AodyoUpdate.STRFTIME_PATTERN, update.releasedate)
		self.update_description = u"%s v%s (%s)" % (update.product, update.version, datestr)
		self.gui.set_product_text(self.update_description)
		self.gui.set_notes(_(fr=update.notes['fr'], en=update.notes['en']))
		self.gui.set_load_btn_text(_(fr=u"Ouvrir une autre mise à jour…", en=u"Open another update…"))
		
		self.dont_look_for_device()

	def onload(self, filename, silent=False):
		if not silent:
			self.update_is_bundled = False

		self.update_filename = filename
		self.root.config(cursor="wait")
		update = None
		try:
			update = UpdateReader(filename)
		except:
			if not silent:
				tkMessageBox.showerror(_(fr=u"Fichier invalide", en=u"Invalid file"), _(fr=u"La mise à jour ne peut pas être installée car elle est invalide ou endommagée. Veuillez réessayer avec un autre fichier.", en=u"The update cannot be installed because it is invalid or damaged. Please try again with another file."), parent=self.root)
			self.root.config(cursor="")
		finally:
			self.root.config(cursor="")
			if update != None:
				if self.update_is_bundled:
					self.root.title(u"%s — %s v%s" % (unicode(AodyoUpdate.TITLE, 'utf-8'), unicode(update.product, 'utf-8'), unicode(update.version, 'utf-8')))
				else:
					self.root.title(u"%s — %s" % (unicode(AodyoUpdate.TITLE, 'utf-8'), os.path.basename(filename)))
				self.load_update_info(update)

	def onupdateclick(self, no_first_time_instructions = False):
		if self.state == UpdateState.LOADED:
			if not no_first_time_instructions:
				if self.needs_firsttime:
					self.first_time_done = winregistry.already_plugged_in([(AodyoUpdate.OLD_VENDOR_ID, AodyoUpdate.OLD_PRODUCT_ID), (AodyoUpdate.VENDOR_ID, AodyoUpdate.PRODUCT_ID)])
			if (no_first_time_instructions) or self.first_time_done:
				self.look_for_device()
			else:
				self.show_first_time_instructions()
		elif self.state == UpdateState.WAITING_FOR_DEVICE:
			self.dont_look_for_device()

	def record_first_time_done(self):
		self.first_time_done = True
		if self.needs_firsttime:
			self.first_time_done = winregistry.already_plugged_in([(AodyoUpdate.OLD_VENDOR_ID, AodyoUpdate.OLD_PRODUCT_ID), (AodyoUpdate.VENDOR_ID, AodyoUpdate.PRODUCT_ID)])

	def nothing_happens_clicked(self):
		if self.state == UpdateState.UPDATING:
			return
		self.dont_look_for_device()
		self.show_first_time_instructions()

	def show_first_time_instructions(self):
		def onfinish(slideshow, success):
			pass
		slideshow_fn = None
		if is_xp():
			slideshow_fn = slideshow.winxp
		elif is_7_or_vista():
			slideshow_fn = slideshow.win7
		if not slideshow_fn is None:
			sshow = slideshow_fn(self.root, _(fr=u"Aodyo Update - Première mise à jour", en=u"Aodyo Update - First-time instructions"), self.update.product, onfinish)
			self.root.withdraw()
			sshow.top.lift()
			self.root.wait_window(sshow.top)
			self.root.deiconify()
		self.record_first_time_done()
		self.onupdateclick(True)

	def look_for_device(self):
		self.state = UpdateState.WAITING_FOR_DEVICE
		self.gui.set_update_btn_enabled(True)
		if self.needs_firsttime:
			self.gui.set_load_btn_text(_(fr=u"Rien ne se passe?", en=u"Nothing seems to happen?"))
			self.gui.set_load_btn_enabled(True)
		else:
			self.gui.set_load_btn_enabled(False)
		self.gui.onloadspecial = self.nothing_happens_clicked
		self.gui.set_update_btn_text(AodyoUpdate.CANCEL_TEXT)
		self.gui.set_info_text(_(fr=u"Branchez directement votre %s sur un port USB de votre ordinateur, puis allumez-le. S'il est déjà branché, éteignez puis rallumez.", en=u"Plug your %s directly to your computer USB port, then turn it on. If already plugged in, turn it off and back on again.") % (self.update.product))
		self.hid = hid.device()
		self.wait_for_device()

	def beforeupdate_info(self):
		return AodyoUpdate.BEFOREUPDATE_INFO % (self.update.product, AodyoUpdate.UPDATE_TEXT)

	def dont_look_for_device(self):
		self.root.config(cursor="")
		self.state = UpdateState.LOADED
		self.gui.set_info_text(self.beforeupdate_info())
		self.gui.set_update_btn_text(AodyoUpdate.UPDATE_TEXT)
		self.gui.set_update_btn_enabled(True)
		self.gui.set_load_btn_text(_(fr=u"Ouvrir une autre mise à jour…", en=u"Open another update…"))
		self.gui.set_load_btn_enabled(True)
		self.gui.onloadspecial = None

	def wait_for_device(self):
		if self.state != UpdateState.WAITING_FOR_DEVICE:
			return

		ok = False
		try:
			self.hid.open(AodyoUpdate.VENDOR_ID, AodyoUpdate.PRODUCT_ID)
			self.device_bl_version = 2
			ok = True
		except IOError as e:
			pass
		if ok:
			self.ondeviceconnected()
		else:
			self.wait_for_old_device()

	def wait_for_old_device(self):
		if self.state != UpdateState.WAITING_FOR_DEVICE:
			return

		ok = False
		try:
			self.hid.open(AodyoUpdate.OLD_VENDOR_ID, AodyoUpdate.OLD_PRODUCT_ID)
			self.device_bl_version = 1
			ok = True
		except IOError as e:
			pass
		if ok:
			self.ondeviceconnected()
		else:
			self.root.after(200, self.wait_for_device)

	def ondeviceconnected(self):
		self.state = UpdateState.UPDATING
		self.root.config(cursor="wait")
		self.gui.set_update_btn_enabled(False)
		self.gui.set_update_btn_enabled(False)

		self.gui.set_load_btn_enabled(False)
		self.gui.set_progress_bar_visible_instead_of_load_btn(True)
		self.gui.set_info_text(_(fr=u"Mise à jour en cours…\nNe retirez pas le câble avant la fin de la procédure.", en=u"Update in progress…\nDo not unplug the cable before the procedure ends."))

		self.root.after(50, self.check_device_and_continue)

	def check_device_and_continue(self):
		ok = False
		device = None
		bootinfo = None
		try:
			self.hid.set_nonblocking(False)
			device = Device(HidApiWrapper(self.hid))
			bootinfo = device.cmd_info()
			ok = bootinfo['McuType'] == 'STM32F4XX'
		except:
			pass

		self.program_device(device, bootinfo) if ok else self.wrap_up_update(False)

	def program_device(self, device, bootinfo):
		self.failure_reason = "(no failure reason)"
		self.programmer = deviceprogrammer.program(device, bootinfo, self.update.firmware)
		self.continue_program_device()

	def continue_program_device(self):
		try:
			progress = self.programmer.next()
			self.progress = progress
			self.gui.set_progress(progress)
			self.root.after(50, self.continue_program_device)
		except StopIteration:
			self.wrap_up_update(True)
		except ExpectException as e:
			self.failure_reason = "USB error (expected %d, got %d) at %d%% with %s, file '%s', bootloader v%d." % (e.expected, e.received, self.progress * 100, self.update_description, self.update_filename, self.device_bl_version)
			self.wrap_up_update(False)
		except:
			# Better traceback
			tb = traceback.extract_tb(sys.exc_traceback)
			self.failure_reason = "%s: %s\n" % (sys.exc_info()[0]().__class__.__name__, sys.exc_info()[1]) + "\n".join(reversed([" - %s.%s:%d" % (os.path.splitext(os.path.basename(e[0]))[0], e[2], e[1]) for e in tb]))
			# Standard traceback
			# self.failure_reason = re.sub(r'File \".+[\\/](\w+)\.py\"', r'', traceback.format_exc())
			self.wrap_up_update(False)

	def wrap_up_update(self, ok):
		self.updateok = ok
		self.root.after(100, self.inform_of_update_result)

	def inform_of_update_result(self):
		ok = self.updateok
		if ok:
			tkMessageBox.showinfo(_(fr=u"Mise à jour effectuée", en=u"Update successful"), _(fr=u"Votre %s a été mis à jour avec succès à la version %s.", en=u"Your %s has been successfully updated to version %s.") % (self.update.product, self.update.version), parent=self.root)
		else:
			tkMessageBox.showerror(_(fr=u"Mise à jour incomplète", en=u"Update failed"), _(fr=u"Votre %s n'a pas été mis à jour en raison d'une erreur. Vous ne pourrez plus l'utiliser correctement avant d'avoir réessayé de faire une mise à jour.\n\nVeuillez l'éteindre, vérifier que la connexion à votre ordinateur est fiable, puis recommencer.\n\nSi le problème persiste ou que vous ne pouvez plus effectuer de mise à jour, veuillez contactez support@aodyo.com en indiquant si possible les informations suivantes:\n\n%s", en=u"Your %s has not been updated due to an error. You will not be able to use it correctly before trying this update again.\n\nPlease turn it off, check that the connection to your computer is reliable, then try again.\n\nIf the problem persists, or if you cannot update your product anymore, please contact support@aodyo.com, providing the following details if possible:\n\n%s") % (self.update.product, self.failure_reason), parent=self.root)
			#tkMessageBox.showerror(_(fr=u"Mise à jour incomplète", en=u"Update failed"), _(fr=u"Votre %s n'a pas été mis à jour en raison d'une erreur.\nVeuillez recommencer et contactez support@aodyo.com si le problème persiste, en indiquant si possible les informations suivantes:\n\n%s", en=u"Your %s has not been updated due to an error.\nPlease try again and contact support@aodyo.com if the problem persists, providing the following details if possible:\n\n%s") % (self.update.product, self.failure_reason), parent=self.root)

		sys.exit(0)
		# self.hid = None
		# self.dont_look_for_device()


if __name__ == '__main__':
	filename = None
	if len(sys.argv) > 1:
		filename = sys.argv[1]
	au = AodyoUpdate(filename)
	au.mainloop()
